"""
Module realizes base plugin class
"""
import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import Generic, TypeVar

from sanic import HTTPResponse
from sanic.response import JSONResponse
from sanic.views import HTTPMethodView

from .plugin_meta import PluginMeta

T_LUNA_APPLICATION = TypeVar("T_LUNA_APPLICATION")


class BasePlugin(ABC, metaclass=PluginMeta):
    """
    Base abstract class.
    Attributes:
        app: running application
    """

    def __init__(self, app: T_LUNA_APPLICATION):
        self.app = app

    @property
    def pluginName(self) -> str:
        """
        Return plugin name

        Returns:
            class name
        """
        return type(self).__name__

    @abstractmethod
    async def initialize(self):
        """
        Initialize plugin.

        Open connections, etc
        """

    @abstractmethod
    async def close(self):
        """
        Stop plugin.

        Close all open connections, etc
        """


class BaseEventPlugin(BasePlugin):
    """
    Base class for plugin which is triggered on an event.
    """

    @property
    @abstractmethod
    def eventName(self) -> str:
        """
        Event name for handling

        Returns:
            event name
        """

    @abstractmethod
    async def handleEvent(self, *args, **kwargs):
        """
        Handle event

        Args:
            *args: positional arg for event handler function
            **kwargs: named arg for event handler function
        """


def pluginHTTPResponse(
    statusCode: int = 200,
    body: bytes | str = None,
    json: dict | list = None,
    headers: dict = None,
    **kwargs,
) -> HTTPResponse:
    """
    Prepare response to be sent back to the client.

    Args:
        statusCode: response status code
        body: pure body
        json: json as object
        headers: response headers

    Returns:
        http response object

    """
    headers = headers or {}

    if json is not None:
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        return JSONResponse(body=json, headers=headers, status=statusCode, **kwargs)

    if "Content-Type" not in headers:
        headers["Content-Type"] = "application/octet-stream"
    return HTTPResponse(body=body, headers=headers, status=statusCode, **kwargs)


class BaseBackgroundHandler(HTTPMethodView):
    """
    Base handler for background process
    """

    response = staticmethod(pluginHTTPResponse)


PluginHandlerType = TypeVar("PluginHandlerType", bound=type[BaseBackgroundHandler] | Callable[..., HTTPResponse])


class BaseBackgroundPlugin(BasePlugin, Generic[PluginHandlerType]):
    """
    Base plugin for background process
    """

    def addRoute(self, route: str, handler: PluginHandlerType, methods: Iterable[str] = None, **kwargs):
        """
        Register a new route.
        """
        if inspect.isfunction(handler):
            routeHandler = handler
        elif inspect.isclass(handler) and issubclass(handler, BaseBackgroundHandler):
            routeHandler = handler.as_view()
        else:
            raise RuntimeError("Unsupported handler type")

        self.app.sanic.router.reset()
        self.app.sanic.add_route(handler=routeHandler, uri=route, methods=methods, **kwargs)
        self.app.sanic.finalize()

    @abstractmethod
    async def start(self):
        """
        Start background process.

        .. warning::
            The function suppose that the process is handle in this coroutine. The coroutine must start
            the process only without awaiting a end of the process
        """
