from typing import Tuple, Union

from sqlalchemy import Column
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.elements import ColumnClause
from vlutils.descriptors.data import DescriptorsEnum

# internal database match function name
DB_MATCH_FUNCTION_NAME = "VLMatch"


class VLMatch(ColumnClause):  # pylint: disable-msg=too-many-ancestors
    """
    VLMatch database function abstraction. Releases descriptor matching interface.

    References:
        https://docs.sqlalchemy.org/en/13/core/compiler.html

    Attributes:
        self.reference (Union[bytes, ColumnClause]): reference descriptor
        self.candidate (Union[bytes, Column]): candidate descriptor or column
        self.descriptorVersion (int): descriptor version

    """

    #: str ALIAS - recommended column name
    ALIAS = "similarity"
    #: str POSTGRES_HEX_MASK - SQL cast from hex to bytes for PostgresDB
    POSTGRES_HEX_MASK = "'\\x{}'::bytea"
    #: str ORACLE_HEX_MASK - SQL cast from hex to bytes for PostgresDB
    ORACLE_HEX_MASK = "HEXTORAW('{}')"

    def __init__(
        self,
        text_: str = DB_MATCH_FUNCTION_NAME,
        type_=None,
        is_literal=False,
        _selectable=None,
        *,
        reference: Union[bytes, ColumnClause] = None,
        candidate: Union[bytes, Column] = None,
        descriptorVersion: int = None,
    ):
        super().__init__(text=text_, type_=type_, is_literal=is_literal, _selectable=_selectable)
        if self.name == DB_MATCH_FUNCTION_NAME:
            self.reference = reference
            self.candidate = candidate
            self.descriptorLength = DescriptorsEnum[f"d{descriptorVersion}"].value.length

    def unique_params(self, *args, **kwargs):  # pylint: disable-msg=W0613
        return super().unique_params()

    def params(self, *args, **kwargs):  # pylint: disable-msg=W0613
        return super().params()


def getMatchItems(vlMatch: VLMatch, hexMask: str) -> Tuple[str, str]:
    """
    Get reference and candidate to compile match object
    Args:
        vlMatch: match object to compile
        hexMask: hex mask

    Returns:
        reference (descriptor) and candidate (descriptor or column name)

    """
    if isinstance(vlMatch.reference, bytes):
        reference = hexMask.format(vlMatch.reference.hex())
    else:
        reference = vlMatch.reference.expression
    if isinstance(vlMatch.candidate, bytes):
        candidate = hexMask.format(vlMatch.candidate.hex())
    else:
        candidate = vlMatch.candidate.expression
    return reference, candidate


@compiles(VLMatch, "postgresql")
def CVLMatch(vlMatch: VLMatch, compiler, **kwargs) -> str:  # pylint: disable-msg=E0102,W0613
    """
    Get compiled VLMatch function.

    Args:
        vlMatch: match object to compile
        compiler: never used
        **kwargs: never used

    Returns:
        func call as SQL statement part

    """
    hexMask = vlMatch.POSTGRES_HEX_MASK
    reference, candidate = getMatchItems(vlMatch, hexMask)
    return f"{vlMatch.name}({reference}, {candidate}, {vlMatch.descriptorLength})"


@compiles(VLMatch, "oracle")
def CVLMatch(vlMatch: VLMatch, compiler, **kwargs) -> str:  # pylint: disable-msg=E0102,W0613
    """
    Get compiled VLMatch function.

    Args:
        vlMatch: match object to compile
        compiler: never used
        **kwargs: never used

    Returns:
        func call as SQL statement part

    """
    # wtf???
    if vlMatch.name != DB_MATCH_FUNCTION_NAME:
        return vlMatch.name
    hexMask = vlMatch.ORACLE_HEX_MASK
    reference, candidate = getMatchItems(vlMatch, hexMask)
    return f"{vlMatch.name}({reference}, {candidate}, {vlMatch.descriptorLength})"
